from __future__ import annotations

import logging
from dataclasses import dataclass
from pathlib import Path

from ckit.config.common import get_global_commands_dir
from ckit.config.config import Config
from ckit.config.yaml_parser import YamlParser
from ckit.group import Group


@dataclass
class ConfigLoader:
    load_local: bool = True
    load_global: bool = True

    def load(self) -> Config:
        """
        Check for the existence of local- and global configuration files, and if they do not exist,
        ask the user if they should be created.
        """
        local_groups = self._load_local() if self.load_local else None
        global_groups = self._load_global() if self.load_global else None

        if not local_groups and not global_groups:
            exit("No configuration files were found. Did you initialize the application with `ckit init`?")

        return Config(local_groups=local_groups, global_groups=global_groups)

    def _load_global(self):
        """
        Find all .yaml files in the global commands directory, and extract the command groups and their commands.
        """
        logging.debug("Loading the global configuration files.")
        global_commands_dir = get_global_commands_dir()
        global_groups = Group()
        if global_commands_dir.exists():
            yaml_files = []
            for extension in ["*.yaml", "*.yml"]:
                yaml_files.extend(global_commands_dir.rglob(extension))
            logging.debug(f"Found the following global command files: {[str(file) for file in yaml_files]}")
            if yaml_files:
                for yaml_file in yaml_files:
                    new_groups = YamlParser().parse(yaml_file)
                    global_groups.join(new_groups)
                return global_groups
        return {}

    def _load_local(self):
        logging.debug("Loading the local configuration files.")
        if Path("ckit.yaml").exists():
            return YamlParser().parse(Path("ckit.yaml"))
        return {}



from __future__ import annotations

from dataclasses import dataclass

from ckit.command import Command


@dataclass
class Group:
    """
    A Group object can hold either:
    - Other Groups (That in turn hold other Groups or Commands)
    - Commands
    """

    elements: dict[str, Group | Command] = None

    def __post_init__(self):
        if not self.elements:
            self.elements = {}

    def get_names(self) -> list[str]:
        return list(self.elements.keys())

    def get(self, name: str) -> Command:
        return self.elements[name]

    def add(self, name: str, element: Group):
        self.elements[name] = element

    def join(self, group: Group):
        self.elements = {**self.elements, **group.elements}

    def contains_only_commands(self):
        return all(isinstance(el, Command) for el in self.elements.values())

    def contains_only_groups(self):
        return all(isinstance(el, Group) for el in self.elements.values())

    def get_names_by_type(self):
        """
        Return a dict with keys 'groups' and 'commands', containing the names of the groups and commands
        respectively.
        """
        return {
            "groups": [k for k, v in self.elements.items() if isinstance(v, Group)],
            "commands": [k for k, v in self.elements.items() if isinstance(v, Command)],
        }

    def __repr__(self):
        return "{" + "\n".join("{!r}: {!r},".format(k, v) for k, v in self.elements.items()) + "}"

    def __str__(self):
        return "{" + "\n".join("{!r}: {!r},".format(k, v) for k, v in self.elements.items()) + "}"



from __future__ import annotations

import logging
import os
from pathlib import Path


def get_global_commands_dir() -> Path:
    """
    Get the directory that contains the global command .yaml files.
    """
    home = Path(os.environ["CKIT_HOME"]) if "CKIT_HOME" in os.environ else Path.home() / "ckit"
    logging.debug(f"The global command directory to source .yaml files from is {home}")
    return home



from __future__ import annotations

import curses
from dataclasses import dataclass, field
from typing import Optional

from blessed import Terminal


@dataclass
class TerminalLine:
    text: str
    selectable: bool
    selected: bool = None

    def select(self):
        self.selected = True

    def deselect(self):
        self.selected = False


@dataclass
class GroupPicker:
    """
    Given some values in groups, pick a value from a group.

    Example:

        group, value = GroupPicker({"fruit": ["apple", "orange", "pear"], "animals": ["dog", "bear"]}).pick()

    Args:
        options: The options to pick from. Should be passed as a dict, where each key is the name of a group, and the
            corresponding value for that entry should be a list of strings.
        title: Optional, title to print aobve the prompt.
    """

    term: Terminal
    options: dict[str, list[str]]
    title: Optional[str] = None
    lines: list[TerminalLine] = field(init=False, default=None)

    def __post_init__(self) -> None:
        # Create the terminal lines
        self.lines = []
        if self.title:
            self.lines += [TerminalLine(text, False) for text in self.title.split("\n") + [""]]
        for group_name, list_of_values in self.options.items():
            self.lines.append(TerminalLine(f"[{group_name}]", False))
            for value in list_of_values:
                self.lines.append(TerminalLine(value, True, False))
            self.lines.append(TerminalLine("", False))

        # Select the first selectable line.
        for line in self.lines:
            if line.selectable:
                line.selected = True
                break

    def select_prev(self):
        """
        Select the previous selectable line.
        """
        selectable_lines = [line for line in self.lines if line.selectable]
        selected_line_index = self.get_index_of_selected_from_selectable()
        to_be_selected_index = (selected_line_index - 1) % len(selectable_lines)
        selectable_lines[selected_line_index].deselect()
        selectable_lines[to_be_selected_index].select()

    def select_next(self):
        """
        Select the next selectable line.
        """
        selectable_lines = [line for line in self.lines if line.selectable]
        selected_line_index = self.get_index_of_selected_from_selectable()
        to_be_selected_index = (selected_line_index + 1) % len(selectable_lines)
        selectable_lines[selected_line_index].deselect()
        selectable_lines[to_be_selected_index].select()

    def get_index_of_selected_from_selectable(self):
        """
        Get index of the currently selected line, relative to the selectable lines. i.e. the non-selectable lines are ignored.
        """
        selectable_lines = [line for line in self.lines if line.selectable]
        selected_line_index = [i for i, x in enumerate(selectable_lines) if x.selected][0]
        return selected_line_index

    def get_index_of_selected(self):
        """
        Get index of the currently selected line, relative to ALL lines.
        """
        selected_line_index = [i for i, x in enumerate(self.lines) if x.selected][0]
        return selected_line_index

    def get_selected(self) -> tuple[str, str]:
        """
        Get the selected group and corresponding value.
        """
        index = self.get_index_of_selected_from_selectable()
        for group, elements in self.options.items():
            if index >= len(elements):
                index -= len(elements)
            else:
                return group, elements[index]

    def pick(self):
        while True:
            # Start at 0,0
            print(self.term.home(), end="")

            # calculate how many lines we should scroll, relative to the top
            max_lines = self.term.height - 1
            index_selected = self.get_index_of_selected()
            scroll_top = 0
            if index_selected >= max_lines:
                scroll_top = index_selected - max_lines + 1

            # Print the lines
            for line in self.lines[scroll_top : scroll_top + max_lines]:
                if line.selectable:
                    if line.selected:
                        print(f"{self.term.clear_eol}> {self.term.orange}{line.text}{self.term.normal}")
                    else:
                        print(f"{self.term.clear_eol}  {line.text}")
                else:
                    print(f"{self.term.clear_eol} {line.text}")

            # Listen for user input
            val = self.term.inkey()
            if val.code == curses.KEY_DOWN:
                self.select_next()
            if val.code == curses.KEY_UP:
                self.select_prev()
            if val.code == curses.KEY_ENTER:
                print(self.term.clear(), end="")
                return self.get_selected()


if __name__ == "__main__":
    k, v = GroupPicker({"fruit": ["apple", "orange", "pear"], "animals": ["dog", "bear"]}).pick()
    print(k, v)



from __future__ import annotations

import os
from contextlib import contextmanager
from pathlib import Path
from typing import Generator


@contextmanager
def run_within_dir(path: Path) -> Generator[None, None, None]:
    """
    Utility function to run some code within a directory, and change back to the current directory afterwards.

    Example usage:

    ```
    with run_within_dir(directory):
        some_code()
    ```

    """
    oldpwd = os.getcwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(oldpwd)



from __future__ import annotations

from dataclasses import dataclass

from ckit.group import Group


@dataclass
class Config:
    local_groups: Group
    global_groups: Group

    def get(self, local_or_global: str):
        if local_or_global == "local":
            return self.local_groups
        elif local_or_global == "global":
            return self.global_groups
        else:
            raise ValueError(f"local_or_global should be either 'local' or 'global', but found {local_or_global}")

    def get_group_names(self):
        """
        Return a dict with the group names per local and global, e.g

        {
            "local" : ["docker", "git"],
            "global" : ["git", "python"]
        }
        """
        result = {}
        if self.local_groups:
            result["local"] = list(self.local_groups.get_names())
        if self.global_groups:
            result["global"] = list(self.global_groups.get_names())
        return result



from __future__ import annotations

from dataclasses import dataclass

from blessed import Terminal

from ckit.config.config import Config
from ckit.group import Group
from ckit.screens.group_picker import GroupPicker
from ckit.screens.picker import Picker


@dataclass
class Core:
    config: Config

    def run(self):
        term = Terminal()
        with term.fullscreen(), term.cbreak(), term.hidden_cursor():
            all_command_group_names = self.config.get_group_names()

            # First time, we have to distinguish between local and global groups.
            local_or_global, command_group_name = GroupPicker(
                term, all_command_group_names, "Please choose a group."
            ).pick()
            group = self.config.get(local_or_global).get(command_group_name)

            # Now, keep picking a Group until a Command is found...
            command = None
            while not command:
                if group.contains_only_commands():
                    choice = Picker(term, group.get_names(), "Please choose a command.").pick()
                    command = group.get(choice)
                elif group.contains_only_groups():
                    choice = Picker(term, group.get_names(), "Please choose a group.").pick()
                    group = group.get(choice)
                else:
                    _, choice = GroupPicker(term, group.get_names_by_type(), "Please make a choice.").pick()
                    selected = group.get(choice)
                    if isinstance(selected, Group):
                        group = selected
                    else:
                        command = selected

        command.run()



from __future__ import annotations

import os
import subprocess
from dataclasses import dataclass

import click


@dataclass
class Argument:
    name: str
    default: str = None


@dataclass
class Boolean:
    """
    Class to hold a boolean option, that will prompt the user for yes/no before executing the command.

    Args:
        name: Name of the boolean. Should match $name in the command.
        prompt: The prompt to show to the user.
        if_true: Value that will replace $name if user chooses `Y`.
        if_true: Value that will replace $name if user chooses `N`. Defaults to "".
        default: Default value (True/False corresponding to `Y`/`N`) if the user presses Enter instead of making a choice.
    """

    name: str
    prompt: str
    if_true: str
    if_false: str = ""
    default: bool = True


class Command:
    def __init__(
        self,
        name: str,
        cmd: str | list(str),
        echo: bool = True,
        args: list[str | dict[str, any]] = None,
        booleans: list[dict[str, any]] = None,
    ) -> None:
        """
        A command object.
        """
        self.name = name
        self.cmd = [cmd] if isinstance(cmd, str) else cmd
        self.echo = echo
        self.arguments = self._parse_arguments(args) if args else None
        self.booleans = self._parse_booleans(booleans) if booleans else None

        if self.arguments:
            self._validate_arguments()

    def _parse_arguments(self, args: list[str | dict[str, any]]) -> list[Argument]:
        """
        Parse a list of argument definitions into Argument objects.

        Args:
            args: A list of either strings, or dictionaries. If the element is a string, it will be an argument without a default.
                If it is a dict, the key will be the name of the argument, and the corresponding value will be the default.
        """
        arguments = []
        for arg in args:
            if isinstance(arg, dict):
                arguments.append(Argument(name=list(arg.keys())[0], default=list(arg.values())[0]))
            elif isinstance(arg, str):
                arguments.append(Argument(name=arg))
            else:
                raise ValueError(f"Argument should be type dict or string, but found {type(arg)}: {arg}")
        return arguments

    def _parse_booleans(self, booleans: list[dict[str, any]]) -> list[Argument]:
        """
        Parse a list of boolean definitions into Boolean objects.

        Args:
            args: A list of dictonaries, where the keys are the names of the Boolean object, and the values are dicts with key-value pairs for the other
            properties. Valid values for these key-value paris are the class attributes of the Boolean class, except for 'name'.
        """
        parsed_booleans = []
        for boolean in booleans:
            name = list(boolean.keys())[0]
            value = list(boolean.values())[0]
            parsed_booleans.append(Boolean(name=name, **value))
        return parsed_booleans

    def _validate_arguments(self):
        """
        Validate that the arguments are actually used in the commands.
        """
        for argument in self.arguments:
            if not any([f"${argument.name}" in command for command in self.cmd]):
                click.echo(
                    f"Argument '{argument.name}' defined for command {self.name}, but '${argument.name}' not found in"
                    " the defined command."
                )

    def run(self):
        cmd = self.cmd

        if self.arguments or self.booleans:
            cmd = self._prompt_and_replace_arguments_and_booleans(cmd)

        for command in cmd:
            if self.echo:
                click.echo(command)
            proc = subprocess.run(self._expand_env_vars(command), shell=True)
            if proc.returncode:
                # Stop running commands if a command fails.
                break
        return proc

    @staticmethod
    def _expand_env_vars(command):
        """
        Replace environment variables with their values.
        """
        return os.path.expandvars(command)

    def _prompt_and_replace_arguments_and_booleans(self, cmd):
        """
        For each argument or boolean, prompt the user for input, and then replace the matching string in the cmd.
        """
        commands_formatted_to_print = "\n".join(cmd)
        click.echo(f"Command{'s' if len(cmd) > 1 else ''} to be run:\n\n{commands_formatted_to_print}\n")
        if self.arguments:
            cmd = self._prompt_and_replace_arguments(cmd)
        if self.booleans:
            cmd = self._prompt_and_replace_booleans(cmd)
        return cmd

    def _prompt_and_replace_arguments(self, cmd):
        """
        For each argument, prompt the user for input, and then replace the matching string in the cmd.
        """
        for argument in self.arguments:
            value = click.prompt(
                f"Please enter a value for argument `{argument.name}`", type=str, default=argument.default
            )
            cmd = [command.replace(f"${argument.name}", value) for command in cmd]
        return cmd

    def _prompt_and_replace_booleans(self, cmd):
        """
        For each boolean, prompt the user for input, and then replace the matching string in the cmd.
        """
        for boolean in self.booleans:
            value = click.confirm(f"{boolean.prompt}", default=boolean.default)
            replacement = boolean.if_true if value else boolean.if_false
            cmd = [command.replace(f"${boolean.name}", replacement) for command in cmd]

        return cmd

    def __repr__(self):
        return f"Command `{self.name}`"

    def __str__(self):
        return f"Command {self.name} ---"



from __future__ import annotations

import importlib.metadata as metadata
import logging

import click

from ckit.config.config_files_initiatior import ConfigFilesInitiator
from ckit.config.config_loader import ConfigLoader
from ckit.core import Core


def configure_logger(_ctx: click.Context, _param: click.Parameter, value: bool) -> None:
    log_level = logging.DEBUG if value else logging.INFO
    logging.basicConfig(level=log_level, handlers=[logging.StreamHandler()], format="%(message)s")


def display_version(ctx: click.Context, _param: click.Parameter, value: bool) -> None:
    if not value or ctx.resilient_parsing:
        return None

    click.echo(f'ckit {metadata.version("ckit")}')  # type: ignore[no-untyped-call]
    ctx.exit()


@click.group(invoke_without_command=True)
@click.option(
    "--local-only",
    "-l",
    is_flag=True,
    help="Boolean flag to only select a command from local files.",
)
@click.option(
    "--global-only",
    "-g",
    is_flag=True,
    help="Boolean flag to only select a command from global files.",
)
@click.option(
    "--verbose",
    "-v",
    is_flag=True,
    help=(
        "Boolean flag for verbosity. Using this flag will display more information about files, imports and"
        " dependencies while running."
    ),
    expose_value=False,
    is_eager=True,
    callback=configure_logger,
)
@click.option(
    "--version",
    is_flag=True,
    is_eager=True,
    expose_value=False,
    callback=display_version,
    help="Display the current version and exit.",
)
@click.pass_context
def ckit(ctx, local_only: bool, global_only: bool) -> None:
    if not ctx.invoked_subcommand:
        if local_only and global_only:
            raise ValueError(
                """Command was run with '--global-only' and '--local-only'. Only one of them can be passed. If you want to load both local and global
            configuration files, simply omit the flags."""
            )
        config = ConfigLoader(load_local=not global_only, load_global=not local_only).load()
        Core(config).run()


@click.option(
    "--download-global-defaults",
    is_flag=True,
    help="""Boolean flag. If set, instead of creating an example ckit.yaml in the global configuration directory, will prompt to download the files
    from the `ckit-files` repository to the global configuration repository instead.""",
)
@click.option(
    "--skip-local",
    is_flag=True,
    help="Boolean flag to indicate that we do not want to initialize a local configuration file.",
)
@click.option(
    "--skip-global",
    is_flag=True,
    help="Boolean flag to indicate that we do not want to initialize global configuration files.",
)
@ckit.command()
def init(download_global_defaults: bool, skip_local: bool, skip_global: bool):
    ConfigFilesInitiator(
        download_global_defaults=download_global_defaults, skip_global=skip_global, skip_local=skip_local
    ).init()



from __future__ import annotations

import logging
from pathlib import Path

import yaml
from yaml.loader import SafeLoader

from ckit.command import Command
from ckit.group import Group


class YamlParser:
    """
    Class to read command groups from a yaml file.
    """

    def __init__(self):
        pass

    def parse(self, file: Path) -> dict[str, Group]:
        with open(file, "rb") as f:
            logging.debug(f"Parsing file {str(file)}")
            specification = yaml.load(f, Loader=SafeLoader)

        if not specification:
            logging.debug("No commands found in.")
            return {}

        return self._parse(specification)

    def _parse(self, specification: dict[str, any]):
        groups = Group()
        for name, element in specification.items():
            if self._is_command(element):
                groups.add(name, Command(name, **element))
            else:
                groups.add(name, self._parse(element))
        return groups

    @staticmethod
    def _is_command(specification: dict[str, any]):
        "Check if 'cmd' is in the specification"
        return "cmd" in specification



from __future__ import annotations

import urllib.request
from dataclasses import dataclass
from pathlib import Path

import click
import requests

from ckit.config.common import get_global_commands_dir

COMMANDS_YAML_DEFAULT = """example:
  echo:
    cmd: "echo Hello World!"

  command-without-echo:
    cmd: "echo Hello World!"
    echo: false

  command-with-env-var:
    cmd: "echo Current user is $USER"

  command-with-user-input:
    cmd: "echo Hello! My name is: $name. My favourite fruit is: $fruit"
    echo: false
    args:
      - name
      - fruit: apple

  command-with-boolean:
    cmd: "ls $detailed"
    echo: false
    booleans:
      - detailed:
          prompt: Show details?
          if_true: -lh

  one-long-command:
    cmd: "echo Lorem ipsum dolor sit amet,
    consectetur adipiscing elit,
    sed do eiusmod tempor incididunt"

  multiple-commands:
    cmd:
      - "echo Lorem ipsum dolor sit amet,"
      - "echo consectetur adipiscing elit"

example-groups:
  group1:
    hello:
      cmd: "echo Hello World from group 1!"
      echo: false
  group2:
    hello:
      cmd: "echo Hello World from group 2!"
      echo: false
  hello:
      cmd: "echo Hello World!"
      echo: false
"""


@dataclass
class ConfigFilesInitiator:
    """
    Args:
      download_global_defaults (bool): If False, simply prompt to create ckit.yaml in the global config directory. If True,
      prompt to download files from the ckit-files repository.
      skip_global: Skip initializing the global configuration directory.
      skip_local: Skip initializing the local configuration file.
    """

    download_global_defaults: bool = False
    skip_global: bool = False
    skip_local: bool = False

    def init(self):
        if not self.skip_global:
            if self.download_global_defaults:
                self._init_global_by_downloading_config_files()
            else:
                self._init_global()
        if not self.skip_local:
            self._init_local()

    def _init_local(self):
        """
        Check if a ckit.yaml file exists in the current directory. Otherwise, ask the user
        if it needs to be created, and if so; created it.
        """
        if Path("ckit.yaml").exists():
            click.echo("A local ckit.yaml already exists.")
        else:
            if click.confirm("Create a ckit.yaml file in the current directory?", default=True):
                with open("ckit.yaml", "w") as f:
                    f.write(COMMANDS_YAML_DEFAULT)

    def _init_global(self):
        """
        Check if there is a global ckit.yaml file. If it does not exist, ask the user if it should be created, and
        if so, create it.
        """
        global_commands_dir = get_global_commands_dir()
        global_commands_file = global_commands_dir / "ckit.yaml"
        if global_commands_file.exists():
            click.echo(f"{global_commands_file} already exists.")
        else:
            if click.confirm(f"Create a global ckit.yaml file in {global_commands_dir}?", default=True):
                global_commands_dir.mkdir(parents=True, exist_ok=True)
                with open(global_commands_file, "w") as f:
                    f.write(COMMANDS_YAML_DEFAULT)

    def _init_global_by_downloading_config_files(self):
        """
        Download config files from the `ckit-files` repository.
        """
        global_commands_dir = get_global_commands_dir()
        config_files = self._list_config_files_from_github()
        if click.confirm(
            f"Download the files {[file['name'] for file in config_files]} and place them in {global_commands_dir}?",
            default=True,
        ):
            global_commands_dir.mkdir(parents=True, exist_ok=True)
            for file in config_files:
                urllib.request.urlretrieve(file["url"], global_commands_dir / file["name"])

    def _list_config_files_from_github(self):
        url = "https://api.github.com/repos/fpgmaas/ckit-files/contents/config-files"
        response_json = requests.get(url).json()
        return [{"name": x["name"], "url": x["download_url"]} for x in response_json]
